# -*- coding: utf-8 -*-
"""
This module contains functions with functionality common for several modules
and helper methods for other modules
"""
# Python module imports
import os
import logging
import sys
from datetime import datetime
import functools

# 3rd party module imports

# Local module imports


# Logger setup
logger = logging.getLogger(__name__)

# Define DEV logging level
logging.DEV = logging.DEBUG - 1

# Define verbose logging level
logging.VERBOSE = logging.DEBUG - 2


def handle_unhandled_exception(exc_type, exc_value, exc_traceback):
    """
    Print stack trace to log-file for unhandled exeptions
    """
    # Do not log program interrupts from keyboard
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    # Print trace to log file
    logger.critical("Unhandled exception", exc_info=(
        exc_type, exc_value, exc_traceback
        ))


def isdate(value):
    """
    Check if `value` is a date on format YYYY-MM-DD.

    Parameters
    ----------
    value : str
        Value to check

    Returns
    -------
    bool
        True if value is a date, otherwise False
    """
    try:
        value = datetime.strptime(value, '%Y-%m-%d')
        return True
    except ValueError:
        return False


def tracelog(func):
    """
    Decorator function for printing tracelogs. Prints timestamp when entering
    function, arguments to function, timestamp when exiting function and return
    value. Also prints docstring line with header "Tracelog".
    Usage: add @tracelog before funktion to decorate. Add "Tracelog: <comment>"
    to function docstring print <comment> from docstring.
    """
    @functools.wraps(func)
    def tracelog_wrapper(*args, **kwargs):

        logger.debug(
            "TRACELOG: %s entered %s() in module %s",
            datetime.now(),
            func.__name__,
            func.__module__,
        )

        logger.verbose(
            "TRACELOG: with parameters %s, %s",
            args,
            kwargs
        )

        if func.__doc__:
            if "Tracelog" in func.__doc__:
                doc_list = [s.strip() for s in func.__doc__.splitlines()]
                logger.debug(
                    "TRACELOG: %s function %s docstring comments: %s",
                    datetime.now(),
                    func.__name__,
                    doc_list[doc_list.index('Tracelog') + 2]
                )

        function_result = func(*args, **kwargs)

        logger.debug(
            "TRACELOG: %s exited %s() in module %s",
            datetime.now(),
            func.__name__,
            func.__module__,
        )

        logger.verbose(
            "TRACELOG: returning %s",
            function_result
        )

        return function_result

    return tracelog_wrapper


def log_level_dev(self, msg, *args, **kwargs):
    """
    Custom log level for development. This log level should only be used locally
    and never be pushed.
    """
    if self.isEnabledFor(logging.DEV):
        self._log(logging.DEV, msg, args, **kwargs)


def log_level_verbose(self, msg, *args, **kwargs):
    """
    Custom log level for verbose messages.
    """
    if self.isEnabledFor(logging.VERBOSE):
        self._log(logging.VERBOSE, msg, args, **kwargs)


def format_verbose_log_header(debug_object, debug_string):
    """
    Construct log-print header

    Parameters
    ----------
    debug_object : object
        The object to print
    debug_string : string
        String with information about the debug print

    Returns
    -------
    -
    """
    debug_header = (
        "\n" +"-" *50 + debug_string +"-" *100 +"\n{0}\n".format(debug_object)
    )

    return debug_header


@tracelog
def get_age_from_id_number(id_number):
    """
    Calculate age from Swedish personnummer/samordningsnummer. The id number
    format supported is yyyymmddnnnc

    Parameters
    ----------
    id_number : str or int
        Id number to use for age calculation. Will be converted to string.
        Supported format is yyyymmddnnnc

    Returns
    -------
    Age calculated from id number.
    """
    id_number = str(id_number)

    year = int(id_number[0:4])
    month = int(id_number[4:6])
    day = int(id_number[6:8])

    # Company
    if month > 12:
        return 0

    # Samordningsnummer has dd+60
    if day > 60:
        day = day - 60

    try:
        date_of_birth = datetime.strptime(
            str(year) + ' '
            + str(month) + ' '
            + str(day), "%Y %m %d"
        ).date()

        age = (
            datetime.now().year - date_of_birth.year
            - ((datetime.now().month, datetime.now().day)
            < (date_of_birth.month, date_of_birth.day))
        )
    except (ValueError) as value_error:
        logger.error("ValueError in calculating age from idnumber: %s", value_error)
        return 0

    return age



# -*- coding: utf-8 -*-
"""
This module contains methods for connecting to API endpoints using the
requests module.
"""
# Python module imports
import logging

# 3rd party module imports
import requests

# Local module imports
from sileopy.common_functions import tracelog


# Logger setup
logger = logging.getLogger(__name__)


@tracelog
def get_request(entity_name, url, api_token):
    """
    Fetch data from API endpoint with a HTTP GET request

    Parameters
    ----------
    entity_name : str
        Name of entity to be fetched from API endpoint
    url : str
        API endpoint URL
    api_token: str
        Token for api-authentication

    Returns
    -------
    Dictionary with the JSON response from the request
    """
    try:
        headers = {'Authorization': api_token}

        http_request = requests.get(
            url,
            headers=headers
        )

        if http_request.status_code == 200:
            logger.info("GET of %s successful.", entity_name)
        else:
            logger.error(
                "GET of %s failed. Error %s",
                entity_name,
                http_request.status_code
            )
            raise requests.ConnectionError("Unexpected HTTP status code.")

        return http_request.json()
    except requests.ConnectionError as connection_error:
        logger.error(
            "Connection error when GETting data from API-endpoint: %s",
            connection_error
        )
        raise connection_error


@tracelog
def post_request(entity_name, url, post_data, api_token):
    """
    Send data to PRIO using a HTTP POST request

    Parameters
    ----------
    entity_name : str
        Name of entity to be fetched from PRIO
    url : str
        API endpoint URL
    post_data : dict
        Dictionary containing data fields to POST in JSON format
    api_token: str
        Token for api-authentication

    Returns
    -------
    Dictionary with the JSON response from the request
    """
    try:
        headers = {'Authorization': api_token}

        http_request = requests.post(
            url,
            data=post_data,
            headers=headers
        )

        if http_request.status_code in (200, 201):
            logger.info("POST of %s successful.", entity_name)
        else:
            logger.error(
                "POST of %s failed. Error: %s",
                entity_name,
                http_request.status_code
            )
            raise requests.ConnectionError("Unexpected HTTP status code.")

        return http_request.json()
    except requests.ConnectionError as connection_error:
        logger.error(
            "Connection error when POSTing data to API-endpoint: %s",
            connection_error
        )
        raise connection_error


@tracelog
def patch_request(entity_name, url, post_data, api_token):
    """
    Send data to PRIO using a HTTP PATCH request

    Parameters
    ----------
    entity_name : str
        Name of entity to be fetched from PRIO
    url : str
        API endpoint URL
    post_data : dict
        Dictionary containing data fields to PATCH in JSON format
    api_token: str
        Token for api-authentication

    Returns
    -------
    Dictionary with the JSON response from the request
    """
    try:
        headers = {'Authorization': api_token}

        http_request = requests.patch(
            url,
            data=post_data,
            headers=headers
        )

        if http_request.status_code in (200, 201):
            logger.info("PATCH of %s successful.", entity_name)
        else:
            logger.error(
                "PATCH of %s failed. Error: %s",
                entity_name,
                http_request.status_code
            )
            raise requests.ConnectionError("Unexpected HTTP status code.")

        return http_request.json()
    except requests.ConnectionError as connection_error:
        logger.error(
            "Connection error when PATCHing data to API-endpoint: %s",
            connection_error
        )
        raise connection_error


@tracelog
def get_prio_version(url, api_token):
    """
    Get current version number of PRIO

    Parameters
    ----------
    url : str
        API endpoint URL
    api_token: str
        Token for api-authentication

    Returns
    -------
    String with the current PRIO version
    """
    try:
        version_data = get_request(
            'prio version'
            , url + 'version/'
            , api_token
        )

        prio_version = version_data['version']

        logger.info("PRIO API Version: %s", prio_version)

        return prio_version

    except requests.ConnectionError as connection_error:
        logger.error("Connection failed: %s", connection_error)


@tracelog
def verify_prio_version(url, api_token, prio_version):
    """
    Check if the current version number of PRIO matches the supported version
    at the caller

    Parameters
    ----------
    url : str
        API endpoint URL
    api_token: str
        Token for api-authentication
    prio_version: str
        Prio version to check

    Returns
    -------
    True if the version number of PRIO matches supported version
    """

    current_prio_version = get_prio_version(url, api_token)

    if current_prio_version == prio_version:
        return True

    return False



# -*- coding: utf-8 -*-
"""
This method contains classes, methods and helper functions for data validation
"""
# Python module imports
from datetime import datetime
import logging
import re

# 3rd party module imports

# Local module imports
from sileopy.common_functions import tracelog


# Logger setup
logger = logging.getLogger(__name__)


@tracelog
def is_amount_above_zero(amount):
    """
    Verify that amount is not zero or below.

    Parameters
    ----------
    amount : Decimal
        Amount to check

    Returns
    -------
    True if `amount` is more than 0.0000
    False if `amount` is 0.0000 or less
    """
    if amount <= 0.0000:
        return False

    return True


@tracelog
def is_date_earlier_than_todays_date(date_to_check):
    """
    Verify that `date_to_check` is before todays date.

    Parameters
    ----------
    date_to_check : str
        Date formatted as string on the format YYYY-MM-DD

    Returns
    -------
    True if `date_to_check` is before todays date
    False if `date_to_check` is today or a future date
    """
    if datetime.now().date() < datetime.strptime(
            date_to_check, '%Y-%m-%d').date():
        return False

    return True


@tracelog
def is_not_empty(value):
    """
    Check if given `value` is not empty

    Parameters
    ----------
    value : any type convertable to str
        Value to check length of

    Returns
    -------
    True if `value` is not blank/empty
    False if `value` is blank/empty
    """
    if not str(value):
        return False

    return True


@tracelog
def is_valid_swedish_id_number(id_number):
    """
    Verify that given id number is on the format yymmdd-nnnc and that the
    checksum digit at the end is correct using the Luhn algoritm.

    Parameters
    ----------
    id_number : int or str
        Id number to validate

    Returns
    -------
    True if `id_number` is a valid Swedish ID number
    False if `id_number` is not a valid Swedish ID number
    """
    id_number = str(id_number)

    #if re.match(r'\d{6}[+-]\d{4}$', id_number) is None:
    if re.match(r'\d{12}$', id_number) is None:
        return False

    id_number = re.sub(r'\D', '', id_number)[2:]
    checksum = sum([int(x) for x in id_number[1::2]])

    for digit in id_number[::2]:
        digit = int(digit)*2
        checksum += digit - 9 if digit > 9 else digit

    if checksum % 10 != 0:
        return False

    return True



# -*- coding: utf-8 -*-
"""
This module fetches new portfolio files from a FTP server to a local directory,
and then moves the files to another directory on the FTP.

Flow:
1. Connect to SFTP server
2. Download files in remote_dir to local_dir
3. Move downloaded files to remote_dir_destination
"""

# Python module imports
import logging
import shutil
import os
from stat import S_ISDIR, S_ISREG
from datetime import datetime

# 3rd party module imports
import paramiko

# Local module imports


# Logger setup
logger = logging.getLogger(__name__)


# All paths to directories must have ending forwardslash
# Always use forwardslash in paths, even in a Windows environment

def is_file(sftp, filename):
    """
    Check if given filename is a file
    """
    try:
        return S_ISREG(sftp.stat(filename).st_mode)
    except IOError:
        return False


def is_dir(sftp, filename):
    """
    Check if given filename is a directory
    """
    try:
        return S_ISDIR(sftp.stat(filename).st_mode)
    except IOError:
        return False


def move_file(path, filename, destination):
    """
    Move file with given `filename` in given `path` to `destination`.

    Parameters
    ----------
    path : str
        Path to directory containing files to be moved
    filename : str
        Filename of file to move
    destination : str
        Destination folder

    Returns
    -------
    True on success
    """
    if destination[-1] not in ['/', '\\']:
        destination = destination + '/'

    # Changing directory facilitates reading files from network paths
    working_dir = os.getcwd()
    os.chdir(os.path.dirname(path))

    if os.path.isdir(destination):
        shutil.move(path + filename, destination + filename)
    else:
        logger.error("File/path error")
        raise FileNotFoundError

    # Return to proper working directory
    os.chdir(working_dir)

    return True


def get_new_ftp_files(
    remote_dir,
    remote_dir_destination,
    local_dir,
    SFTP_HOST,
    SFTP_USER,
    SFTP_PASSWORD,
    SFTP_PORT
    ):
    """
    Download new files from remote_dir at SFTP_HOST to local_dir.
    Files downloaded will be moved remotely to remote_dir_destination

    Parameters
    ----------
    remote_dir : str
        Path to directory containing files to be downloaded
    remote_dir_destination : str
        Destination directory at SFTP_HOST
    local_dir : str
        Local destination directory
    SFTP_HOST : str
        Hostname of the FTP-server
    SFTP_USER : str
        Username of the FTP-account
    SFTP_PASSWORD : str
        Password of the FTP-account

    Returns
    -------
    List of fetched files on success, otherwise False
    """
    # Check FTP-settings
    if not all([SFTP_HOST, SFTP_USER, SFTP_PASSWORD, SFTP_PORT]):
        logger.error("Cannot connect to FTP-server: Missing FTP-settings.")
        return False

    fetched_files = []
    passw = SFTP_PASSWORD

    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    logger.info("Connecting to %s as %s", SFTP_HOST, SFTP_USER)
    ssh.connect(SFTP_HOST, username=SFTP_USER, password=passw, port=SFTP_PORT)
    sftp = ssh.open_sftp()

    sftp.chdir(remote_dir_destination)
    existing_files = sftp.listdir()
    timestamp = datetime.strftime(
            datetime.now(),
            '%Y-%m-%d-%H%M%S'
        )

    sftp.chdir(remote_dir)
    files = sftp.listdir()

    logger.debug("New files to get: %s", files)
    
    for remote_file in files:
        if is_file(sftp, remote_file):
            # Fetch the file to the local directory
            logger.info("Downloading %s", remote_file)
            sftp.get(remote_file, local_dir + remote_file)

            fetched_files.append(remote_file)

            # Move the remote file after fetch, change filename if file exist
            logger.info("Moving " + remote_file + " on server to " + remote_dir_destination)
            if remote_file in existing_files:
                sftp.rename(remote_dir + remote_file, remote_dir_destination + remote_file + '.' + timestamp)
            else:
                sftp.rename(remote_dir + remote_file, remote_dir_destination + remote_file)

    logger.info("Closing connection")
    sftp.close()
    ssh.close()

    return fetched_files



# -*- coding: utf-8 -*-
"""
This method contains classes, methods and helper functions for file handling.
"""
# Python module imports
import logging
import hashlib
import os
import shutil
import platform
from datetime import datetime

# Local module imports
from sileopy.common_functions import tracelog


# Logger setup
logger = logging.getLogger(__name__)


@tracelog
def get_csv_files_in_dir(path):
    """
    Scan given `path` for files with extension .csv and return list with
    file names of any files found

    Parameters
    ----------
    path : str
        Directory path to scan for CSV files

    Returns
    -------
    List of filenames found in given `path`
    """
    csv_files = []

    for entry in os.scandir(path):
        if entry.is_file():
            if entry.name[-4:] == '.csv':
                csv_files.append(entry.name)

    return csv_files


@tracelog
def get_sha256_file_hash(filename):
    """
    Calculate SHA-2 hash for given file.

    Parameters
    ----------
    filename: str
        Filename of the file to hash

    Returns
    -------
    String containing SHA-256 hash
    """
    algo = hashlib.sha256()

    with open(filename, 'rb') as file_to_hash:
        for chunk in iter(lambda: file_to_hash.read(128 * algo.block_size), b''):
            algo.update(chunk)

    return algo.hexdigest()


@tracelog
def move_file(path, filename, destination):
    """
    Move file with given `filename` in given `path` to `destination`.

    Parameters
    ----------
    path : str
        Path to directory containing files to be moved
    filename : str
        Filename of file to move
    destination : str
        Destination folder

    Returns
    -------
    True on success
    """
    if destination[-1] not in ['/', '\\']:
        destination = destination + '/'

    # Changing directory facilitates reading files from network paths
    working_dir = os.getcwd()
    os.chdir(os.path.dirname(path))

    if os.path.isdir(destination):
        shutil.move(path + filename, destination + filename)
    else:
        logger.error("File/path error")
        raise FileNotFoundError

    # Return to proper working directory
    os.chdir(working_dir)

    return True


@tracelog
def move_files(path, filenames, destination):
    """
    Move files with given `filenames` in given `path` to `destination`.

    Parameters
    ----------
    path : str
        Path to directory containing files to be moved
    filenames : list of str
        Filename of file to move
    destination : str
        Destination folder

    Returns
    -------
    True on success
    """
    for filename in filenames:
        move_file(path, filename, destination)

    return True


@tracelog
def adjust_path_slashes(filepath):
    """
    Replace slashes with backslashes or vice versa depending on operating system
    and return the updated file path

    Parameters
    ----------
    filepath : str
        Path to a directory

    Returns
    -------
    new_filepath : str
        The updated path
    """
    if platform.system() == 'Windows':
        new_filepath = filepath.replace('/', '\\')
    else:
        new_filepath = filepath.replace('\\', '/')
    return new_filepath


@tracelog
def get_file_detail(filepath, filename, entity, origin):
    """
    Hash input file and return file details, i.e. hash-value, file name,
    origin and entity

    Parameters
    ----------
    filepath : string
        path to file to hash
    filename : string
        name of the file
    entity : string
        entity of the file
    origin : string
        origin of the file

    Returns
    -------
    file_detail : dict
        Dictionary with file details
    """
    hash_value = get_sha256_file_hash(filepath)
    file_detail = {
        'filename': filename,
        'entity': entity,
        'filehash': hash_value,
        'origin': origin
    }
    return file_detail


@tracelog
def get_portfoliodatafile_prefix(creditor_name):
    """
    Create a string containing creditor name, current year, month, day and time.

    Parameters
    ----------
    creditor_name : str
        Name of the creditor

    Returns
    -------
    portfoliodatafile_prefix : str
        The prefix for the portfolio data file. 
    """
    datestamp = datetime.strftime(datetime.now(), '%Y-%m-%d-%H%M%S%f')[:-2]

    portfoliodatafile_prefix = 'portfolio_' + datestamp + '_' \
        + creditor_name.replace(' ', '_') + '_'
    
    return portfoliodatafile_prefix


# -*- coding: utf-8 -*-
"""
This method contains classes, methods and helper functions for Opsgenie
"""
# Python module imports
import opsgenie_sdk
import logging

# Local module imports
from sileopy.common_functions import tracelog


# Logger setup
logger = logging.getLogger(__name__)


class Opsgenie:
    """
    Main class for Opsgenie
    """
    @tracelog
    def __init__(self, key, host):
        self.conf = self.conf = opsgenie_sdk.configuration.Configuration()
        self.conf.api_key['Authorization'] = key
        self.conf.host = host

        self.api_client = opsgenie_sdk.api_client.ApiClient(configuration=self.conf)
        self.alert_api = opsgenie_sdk.AlertApi(api_client=self.api_client)
        self.heartbeat_api = opsgenie_sdk.HeartbeatApi(api_client=self.api_client)
        self.heartbeat_name = ''

    @tracelog
    def create_alert(
        self,
        alert_message,
        alert_description,
        alert_priority,
        alert_script_path,
        alert_script_hostname, 
        alert_alias = ''
        ):
        """
        Create an alert and send to Opsgenie

        Parameters
        ----------
        self : Opsgenie
            The object itself
        alert_message : String
            The message/subject of the alert
        alert_description : String
            The description of the alert
        alert_priority : String
            The priority of the alert (P1-P5)       
        alert_script_path : String
            The path to the script sending the alert
        alert_script_hostname : String
            The hostname or container ID running the script
        alert_alias : String
            The alias of the alert
            
        Returns
        -------
        -
        """
        body = opsgenie_sdk.CreateAlertPayload(
            message=alert_message,
            description=alert_description or '',
            details={
                'Alert script path': alert_script_path \
                    or '</path/to/alert script>',
                'Alert script hostname / container ID': alert_script_hostname \
                    or '<name of host or ID of container running the alert script>'
            },
            priority=alert_priority, 
            alias=alert_alias or ''
        )
        logger.debug('create_alert body: \n%s', body)
        try:
            response = self.alert_api.create_alert(create_alert_payload=body)
            logger.info(response)
            return response
        except opsgenie_sdk.ApiException as err:
            logger.error(
                "Exception when calling AlertApi->create_alert: %s\n" % err
            )
    
    @tracelog
    def create_heartbeat_ping(self):
        """
        Create a heartbeat and send to Opsgenie

        Parameters
        ----------
        self : Opsgenie
            The object itself

        Returns
        -------
        -
        """
        try:
            logger.debug('Sending ping to heartbeat name: %s', self.heartbeat_name)
            ping_response = self.heartbeat_api.ping(self.heartbeat_name)
            logger.info(ping_response)
            return ping_response
        except opsgenie_sdk.ApiException as err:
            logger.error("Exception when calling HeartBeatApi->ping: %s\n" % err)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sileopy']

package_data = \
{'': ['*']}

install_requires = \
['opsgenie-sdk>=2.1.5,<3.0.0',
 'paramiko>=2.11.0,<3.0.0',
 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'sileopy',
    'version': '0.1.0',
    'description': 'Ett bibliotek för python-moduler med allmän funktionalitet som kan vara användbar i flera projekt.',
    'long_description': None,
    'author': 'Johan Magnusson',
    'author_email': 'johan.magnusson@sileokapital.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
